/*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 
*/


#include <WCore>

#include "RBoard.h"
#include "RFBoard.h"
#include "RPlayer.h"




void cRPlayer::
TestPlay(long maxtestdepth, long maxtestrounds)
{
	cReversiBoard rb;
	rb.Reset();
	bool fTerm=false;

	long j;
	cReversiBoard::Piece p = cReversiBoard::kPiece1;
	for (j=0; j<maxtestrounds;j++)
	{
		cReversiBoard::Position x,y;

		cRPlayer(Computer, p).Think(&rb, maxtestdepth, &x, &y, 0, &fTerm);
		rb.MakeMove(p, x,y);
		p = cReversiBoard::OtherPiece(p);
	}
/*

	this->MakeMove(kPiece1, 4,5);
	this->MakeMove(kPiece2, 5,5);
	this->MakeMove(kPiece1, 6,5);
	this->MakeMove(kPiece2, 3,5);
	this->MakeMove(kPiece1, 2,5);
	this->MakeMove(kPiece2, 3,6);
	this->MakeMove(kPiece1, 2,2);
	this->MakeMove(kPiece2, 7,5);
	this->MakeMove(kPiece1, 3,7);
	this->MakeMove(kPiece2, 1,1);
*/

/*
	0x0000
	0x2000
	0x0800
	0x0240
	0x0180
	0x05aa
	0x0100
	0x0100

	Mobility:
	0x00
	0x00
	0x10
	0x24
	0x04
	0x00
	0x0c
	0x00
*/
}


void cRPlayer::
StopThinking()
{
	if (this->ptThinker)
		DestroyThread((cThreadTask**)&this->ptThinker);
}

void cRPlayer::
StartThinking(const cReversiBoard * prb, cReversiBoard::Position *px, cReversiBoard::Position *py, cTask *pTrigger)
{
	long maxdepth=this->maxdepth;

	this->StopThinking();
	Assert(!this->ptThinker);

	if ((prb->CountPieces() + this->endgameThreshold) >= 64)
		maxdepth *=3;

	this->ptThinker = new cThinkTask(this, prb, maxdepth, px, py, pTrigger);
	if (this->ptThinker)
		this->ptThinker->Init();
}

void cRPlayer::cThinkTask::
Perform()
{
	Assert(this->mommy);
	Assert(this->pBoard);
	Assert(this->px);
	Assert(this->py);
	Assert(this->pTrigger);

	this->mommy->Think(this->pBoard, this->maxdepth, this->px, this->py, this->pTrigger, this->GetDeathWatch());
}

void cRPlayer::
Think(const cReversiBoard * prfb, long maxdepth, cReversiBoard::Position *px, cReversiBoard::Position *py, cTask *pTrigger, const bool *pfTerminate) const
{
	long movevalue;
	ulong ulMinmaxStart = cRFastBoard::gMinmaxMoveCounterTotal;
	cReversiBoard::Position x,y;
	cRFastBoard originalBoard(*prfb);

	Assert(px);
	Assert(py);
	Assert(pfTerminate);

#ifdef DUMPANALYSIS
	ulong tickStart;
	if (cRFastBoard::fDumpMoveAnalysis)
	{
		cRFastBoard::ofsDump	<< "Move Analysis for: " << ((cReversiBoard::kPiece1 == piece)?"Black":"White")
								<< ". depth=" << maxdepth
								<< endl
								<< "B @:" << originalBoard.CountPieces(cReversiBoard::kPiece1)
								<<"\tW 0:"<< originalBoard.CountPieces(cReversiBoard::kPiece2)
#ifdef DEBUG
								<< '(' << originalBoard.FullLineCount() << " full lines." << originalBoard.EmptyLineCount() << " empty lines.)"
#endif
								<< endl
								<< originalBoard;
		tickStart = AGetTicks();
	}
#endif

	switch(this->brain)
	{
		case Loser: originalBoard.pStrengthFunc = &cRFastBoard :: LosingStrength; break;
		case Greedy: originalBoard.pStrengthFunc = &cRFastBoard :: GreedyStrength; break;
		case Positional: originalBoard.pStrengthFunc = &cRFastBoard :: PositionalStrength; break;
		case Mobility: originalBoard.pStrengthFunc = &cRFastBoard :: Strength; break;
		default: Assert(false); break;
	}

	originalBoard.UpdateMobility(piece);
	movevalue = -cRFastBoard::kInfinity;
	x = -1;
	y = 0;
	while(!*pfTerminate && originalBoard.NextValidMove(this->piece , &x, &y))
	{
		cRFastBoard tBoard;
		long newvalue;

		cRFastBoard::gMinmaxMoveCounterTotal++;

		tBoard = originalBoard;
		tBoard.MakeMove(piece, x, y);
		newvalue = -tBoard.ABPrune(cReversiBoard::OtherPiece(this->piece), maxdepth, -movevalue, -cRFastBoard::kInfinity, pfTerminate);

		if (newvalue > movevalue)
		{
			movevalue = newvalue;
			*px = x;
			*py = y;
		}

#ifdef DUMPANALYSIS
		if (cRFastBoard::fDumpMoveAnalysis)
			cRFastBoard::ofsDump <<'['<< char(uint('a')+x)<<long(y+1) << ':' << newvalue <<']';
#endif
	}

#ifdef DUMPANALYSIS
	if (cRFastBoard::fDumpMoveAnalysis)
	{
		ulong tickEnd = AGetTicks();

		cRFastBoard::ofsDump	<< endl
								<< "Moving :" << char(uint('a')+*px) << long(*py+1)
								<< ". Searched " << cRFastBoard::gMinmaxMoveCounterTotal - ulMinmaxStart
								<< " boards at "
								<< (((double)(cRFastBoard::gMinmaxMoveCounterTotal - ulMinmaxStart)*kATicksPerSecond)/(double)AMax((ulong)1,tickEnd-tickStart))
								<< " per second."
								<< endl << endl;
	}

#endif
	Assert(*px<8);
	Assert(*py<8);

	if(pTrigger && !(*pfTerminate))
		GAppl()->AddEvent(pTrigger);
}


